#!/usr/bin/perl -w

############################################################################
#                                                                          #
#  c++2ps a pretty print converter for C++                                 #
#                                                                          #
#  date         : 2000, Sep 10                                             #
#  release      : 0.1                                                      #
#  release_date : 2000, Sep 10                                             #
#  file         : c++2ps                                                   #
#  author       : Enrico Bertolazzi                                        #
#  email        : enrico.bertolazzi@ing.unitn.it                           #
#                                                                          #
#  This program is free software; you can redistribute it and/or modify    #
#  it under the terms of the GNU General Public License as published by    #
#  the Free Software Foundation; either version 2, or (at your option)     #
#  any later version.                                                      #
#                                                                          #
#  This program is distributed in the hope that it will be useful,         #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of          #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
#  GNU General Public License for more details.                            #
#                                                                          #
#  You should have received a copy of the GNU General Public License       #   
#  along with this program; if not, write to the Free Software             #
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               #
#                                                                          #
#  Copyright (C) 2000                                                      #
#                                                                          #
#           ___    ____  ___   _   _        ___    ____  ___   _   _       #
#          /   \  /     /   \  \  /        /   \  /     /   \  \  /        #
#         /____/ /__   /____/   \/        /____/ /__   /____/   \/         #
#        /   \  /     /  \      /        /   \  /     /  \      /          #
#       /____/ /____ /    \    /        /____/ /____ /    \    /           #
#                                                                          #
#      Enrico Bertolazzi                                                   #
#      Dipartimento di Ingegneria Meccanica e Strutturale                  #
#      Universita` degli Studi di Trento                                   #
#      Via Mesiano 77, I-38050 Trento, Italy                               #
#                                                                          #
############################################################################

use Getopt::Long;

my $endl = "\n";

$tablen = 8;

#
# Available colors
#
# BLACK, GRAY, MAROON, PURPLE, GREEN, OLIVE, NAVY
# TEAL, SILVER, WHITE, RED, FUCHSIA, LIME, YELLOW
# BLUE, AQUA
#

$text_color         = "BLACK";
$comment_color      = "RED";
$keyword_color      = "BLUE";
$function_color     = "MAROON";
$preprocess_color   = "OLIVE";
$number_color       = "RED";
$string_color       = "GRAY";
$class_color        = "NAVY";
$alias_color        = "TEAL";
$braket_color       = "AQUA";
$symbol_color       = "LIME";
$macro_color        = "PURPLE";

#$text_font          = "\\textrm";
$comment_font       = "";
$keyword_font       = "";
$function_font      = "";
$preprocess_font    = "";
$number_font        = "";
$string_font        = "";
$class_font         = "";
$alias_font         = "";
$braket_font        = "";
##$symbol_font        = "";
$macro_font         = "";

$verbose            = 0;
$silent             = 0;
$columns            = 1;
@port_land          = ("", "portrait", "landscape");
@main_font_size     = ("", "", "\\small");

$rule_width         = "0.5pt";

@begin_multicols    = ("", "", "\\columnsep=2em\\relax" . $endl .
                               "\\columnseprule=${rule_width}\\relax" . $endl .
                               "\\begin{multicols\*}{2}" . $endl);
@end_multicols      = ("", "", "\\end{multicols\*}");

@margin             = ("", "{1.7cm,1.7cm}", "{1.4cm,1.4cm}");

## PREPROCESSING DIRECTIVE
$PREPROCESSING      = 'include' . '|' .
                      'define'  . '|' .
                      'undef'   . '|' .
                      'ifdef'   . '|' .
                      'ifndef'  . '|' .
                      'else'    . '|' .
                      'elif'    . '|' .
                      'endif'   . '|' .
                      'if';

## KEYWORDS
$KEYWORDS =           'bool'             . '|' .
                      'true'             . '|' .
                      'false'            . '|' .
                      'char'             . '|' .
                      'short'            . '|' .
                      'int'              . '|' .
                      'long'             . '|' .
                      'unsigned'         . '|' .
                      'signed'           . '|' .
                      'float'            . '|' .
                      'double'           . '|' .
                      'enum'             . '|' .

                      'if'               . '|' .
                      'else'             . '|' .
                      'for'              . '|' .
                      'do'               . '|' .
                      'while'            . '|' .
                      'switch'           . '|' .
                      'case'             . '|' .
                      'break'            . '|' .
                      'default'          . '|' .
                      'goto'             . '|' .
                      'continue'         . '|' .

                      'return'           . '|' .

                      'auto'             . '|' .

                      'static'           . '|' .
                      'register'         . '|' .
                      'volatile'         . '|' .

                      'struct'           . '|' .
                      'union'            . '|' .

                      'typedef'          . '|' .
                      'defined'          . '|' .

                      'const'            . '|' .

                      'extern'           . '|' .
                      'void'             . '|' .
                      'sizeof'           . '|' .

                      'new'              . '|' .
                      'delete'           . '|' .

                      'this'             . '|' .
                      'operator'         . '|' .
                      'class'            . '|' .
                      'public'           . '|' .
                      'protected'        . '|' .
                      'private'          . '|' .
                      'virtual'          . '|' .
                      'friend'           . '|' .
                      'inline'           . '|' .

                      'template'         . '|' .
                      'typename'         . '|' .
                      'namespace'        . '|' .
                      'using'            . '|' .
                      'explicit'         . '|' .

                      'try'              . '|' .
                      'catch'            . '|' .
                      'throw'            . '|' .

                      'const_cast'       . '|' .
                      'static_cast'      . '|' .
                      'dynamic_cast'     . '|' .
                      'reinterpret_cast' . '|' .
                      'mutable'          . '|' .

                      'overload';

$FUNCTIONS =          'sin'       . '|' .
                      'cos'       . '|' .
                      'tan'       . '|' .

                      'asin'      . '|' .
                      'acos'      . '|' .
                      'atan'      . '|' .
                      'atan2'     . '|' .

                      'sinh'      . '|' .
                      'cosh'      . '|' .
                      'tanh'      . '|' .

                      'ceil'      . '|' .
                      'exp'       . '|' .
                      'abs'       . '|' .
                      'floor'     . '|' .
                      'mod'       . '|' .
                      'log'       . '|' .
                      'log10'     . '|' .
                      'pow'       . '|' .
                      'sqrt';

##############################
##############################

&do_conversion;

sub do_conversion {

  GetOptions ('help'               => \$do_help,
              'verbose'            => \$verbose,
              'silent'             => \$silent,
              '2'                  => \$two_column,
              'r'                  => \$do_reverse,
              'tex'                => \$leave_tex,
              
              'text_color=s'       => \$text_color,
              'comment_color=s'    => \$comment_color,
              'keyword_color=s'    => \$keyword_color,
              'function_color=s'   => \$function_color,
              'preprocess_color=s' => \$preprocess_color,
              'number_color=s'     => \$number_color,
              'string_color=s'     => \$string_color,
              'class_color=s'      => \$class_color,
              'alias_color=s'      => \$alias_color,
              'braket_color=s'     => \$braket_color,
              'symbol_color=s'     => \$symbol_color,
              'macro_color=s'      => \$macro_color,

              'comment_font=s'     => \$comment_font,
              'keyword_font=s'     => \$keyword_font,
              'function_font=s'    => \$function_font,
              'preprocess_font=s'  => \$preprocess_font,
              'number_font=s'      => \$number_font,
              'string_font=s'      => \$string_font,
              'class_font=s'       => \$class_font,
              'alias_font=s'       => \$alias_font,
              'braket_font=s'      => \$braket_font,
              #'symbol_font=s'      => \$symbol_font,
              'macro_font=s'       => \$macro_font,
              
              'tab=i'              => \$tablen,
             ) or &do_err_message;
             
  &do_err_message if $do_help;

  $text_color       =~ tr/[a-z]/[A-Z]/;
  $comment_color    =~ tr/[a-z]/[A-Z]/;
  $keyword_color    =~ tr/[a-z]/[A-Z]/;
  $function_color   =~ tr/[a-z]/[A-Z]/;
  $preprocess_color =~ tr/[a-z]/[A-Z]/;
  $number_color     =~ tr/[a-z]/[A-Z]/;
  $string_color     =~ tr/[a-z]/[A-Z]/;
  $class_color      =~ tr/[a-z]/[A-Z]/;
  $alias_color      =~ tr/[a-z]/[A-Z]/;
  $braket_color     =~ tr/[a-z]/[A-Z]/;
  $symbol_color     =~ tr/[a-z]/[A-Z]/;
  $macro_color      =~ tr/[a-z]/[A-Z]/;

  $silent  = 0 if $verbose;
  $columns = 2 if $two_column;

  my $nfile = 0;
  my $i;
  
  foreach $file (@ARGV) {

    if ( -d $file ) {
      print 'skipping directory: ', $file, $endl unless $silent;
    } else {
      if ( $file =~ /\.(c|cc|cpp|C|c\+\+|h|hh|h\+\+|hxx)$/ ) {
        $FILE_LIST[$nfile] = $file;
        &do_encoding($file);
        $nfile++;
      } else {
        print 'skipping file: ',$file, $endl unless $silent;
      }
    }
  }
  
  for ( $i = 0; $i < $nfile; $i++ ) {
    &do_more_encoding($FILE_LIST[$i]);
  }

  &do_userkey_encoding;

  for ( $i = 0; $i < $nfile; $i++ ) {
    &do_decoding($FILE_LIST[$i]);
  }
  
  print "All conversion done.", $endl unless $silent;
}

sub do_err_message {

print <<MSG;
Usage: c++2ps [options] file1 file1 ...

   --help             print this message
   --verbose          more messages
   --silent           no messages
   --2                two column formatting
   --r                pages in reverse order 
   --tab=i            tab length
   --tex              do not remove generated tex file 

   --XXX_color=COLOR
   --XXX_font=FONT
  
XXX   = comment, keyword, function, preprocess, number, string,
        class, alias, braket, macro
COLOR = black, gray, maroon, purple, green, olive, navy, teal,
        silver, white, red, fuchsia, lime, yellow, blue, aqua.
FONT  = textrm, textsf, texttt, textmd, textbf, textup, textit,
        textsl, testsc, emph

by Enrico Bertolazzi (c)
MSG
exit(0);
}

sub do_rgb {
  local ($COLOR,$rgb) = @_;
  my $out = "\\definecolor{${COLOR}}{rgb}{";
  foreach (split(',',$rgb)) {
    $out .= hex($_)/256 . ",";
  }
  $out =~ s/,$/\}/;
  $out;
}

sub do_latex_command {
  local ($CMD,$n,$FONT,$COLOR,$BODY) = @_;
  my $out = "\\providecommand{\\${CMD}}[${n}] " .
            "{{${FONT}{\\color{${COLOR}}${BODY}}}}" .
            $endl;
  $out;
}

sub do_header {
  local ($file) = @_;
  my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday) = gmtime(time);
  $year = $year + 1900;
  $mon  = ("Jan", "Feb", "Mar", "Apr", "May",
           "Jun", "Jul", "Aug", "Oct", "Nov", "Dec")[$mon];
  $mday = "0".$mday if $mday < 10;
  $hour = "0".$hour if $hour < 10;
  $min  = "0".$min  if $min  < 10;

  "\\documentclass[10pt]{article}" . $endl .
  "\\usepackage{geometry,multicol,color,fancyhdr}" . $endl .
  &do_latex_command("KEYWORD", "1",$keyword_font,   $keyword_color,   "\#1") .
  &do_latex_command("FUNCTION","1",$function_font,  $function_color,  "\#1") .
  &do_latex_command("PREPRO",  "1",$preprocess_font,$preprocess_color,"\#1") .
  &do_latex_command("CSTRING", "1",$string_font,    $string_color,    "\#1") .
  &do_latex_command("NUMBER",  "1",$number_font,    $number_color,    "\#1") .
  &do_latex_command("COMMENT", "1",$comment_font,   $comment_color,   "\#1") .
  &do_latex_command("CLASS",   "1",$class_font,     $class_color,     "\#1") .
  &do_latex_command("ALIAS",   "1",$alias_font,     $alias_color,     "\#1") .
  &do_latex_command("MACRO",   "1",$macro_font,     $macro_color,     "\#1") .
  &do_latex_command("CLASS",   "1",$class_font,     $class_color,     "\#1") .
  
  &do_latex_command("FLOAT", "2",$number_font,  $number_color,
                    "\#1\$\\times10\^{\#2}\$") .

  &do_latex_command("PTR",    "0","", $symbol_color,"\$\\rightarrow\$") .
  &do_latex_command("PTRSTAR","0","", $symbol_color,"\$\\rightarrow\\\!\\\!\\ast\$") .
  &do_latex_command("DOTSTAR","0","", $symbol_color,"\$\\ast\$") .
  &do_latex_command("GT",     "0","", $symbol_color,"\$>\$") .
  &do_latex_command("LT",     "0","", $symbol_color,"\$<\$") .
  &do_latex_command("GG",     "0","", $symbol_color,"\$\\gg\$") .
  &do_latex_command("LL",     "0","", $symbol_color,"\$\\ll\$") .
  &do_latex_command("GE",     "0","", $symbol_color,"\$\\geq\$") .
  &do_latex_command("LE",     "0","", $symbol_color,"\$\\leq\$") .
  &do_latex_command("NE",     "0","", $symbol_color,"\$\\neq\$") .
  &do_latex_command("BRA",    "1",$braket_font, $braket_color,"\#1") .

  "\\providecommand{\\EOL}{\\hfill\\BS{}}"   . $endl .
  "\\providecommand{\\BS} {\$\\backslash\$}" . $endl .

  &do_rgb("BLACK","00,00,00") . $endl .
  &do_rgb("GRAY","80,80,80") . $endl .
  &do_rgb("MAROON","80,00,00") . $endl .
  &do_rgb("PURPLE","80,00,80") . $endl .
  &do_rgb("GREEN","00,80,00") . $endl .
  &do_rgb("OLIVE","80,80,00") . $endl .
  &do_rgb("NAVY","00,00,80") . $endl .
  &do_rgb("TEAL","00,80,80") . $endl .
  &do_rgb("SILVER","C0,C0,C0") . $endl .
  &do_rgb("WHITE","FF,FF,FF") . $endl .
  &do_rgb("RED","FF,00,00") . $endl .
  &do_rgb("FUCHSIA","FF,00,FF") . $endl .
  &do_rgb("LIME","00,FF,00") . $endl .
  &do_rgb("YELLOW","FF,FF,00") . $endl .
  &do_rgb("BLUE","00,00,FF") . $endl .
  &do_rgb("AQUA","00,FF,FF") . $endl .
  "\\parindent=0pt\\relax" . $endl .
  "\\renewcommand{\\baselinestretch}{0.85}" . $endl .
  "\\renewcommand{\\headrulewidth}{${rule_width}}" . $endl .
  "\\renewcommand{\\footrulewidth}{${rule_width}}" . $endl .
  "\\geometry{a4paper,headsep=5pt,footskip=18pt,$port_land[$columns],margin=$margin[$columns]}" . $endl .
  "\\pagestyle{fancy}" . $endl .
  "\\lhead{\\textbf{" . &latex_symbols($file) . "}}" . $endl .
  "\\chead{\\textbf{$mon $mday\$^{th}\$ $year}}" . $endl .
  "\\rhead{\\textbf{$hour\$:\$$min}}" . $endl .
  "\\cfoot{\\textbf{page \\thepage{} of \\pageref{LASTPAGE}}}" . $endl .
  "\\begin\{document\}" . $endl .
  "\\ttfamily" . $endl .
  "$begin_multicols[$columns]" . $endl .
  "$main_font_size[$columns]" . $endl;
}

sub do_footer {
  "\\refstepcounter{equation}\\label{LASTPAGE}" . $endl .
  "$end_multicols[$columns]" . $endl .
  "\\end{document}";
}


###############################################
###############################################

sub latex_chars {
  local ($_) = @_;
  
  # coding {}
  s/\{/\@OB\@/g;
  s/\}/\@CB\@/g;
  
  # coding \\
  s/\\/\\BS\{\}/g;

  # coding specials
  s/((\_|\&|\#|\$|\%))/\\$1/g;
  
  # coding symbols
  s/\^/\$\\wedge\$\{\}/g;
  s/\~/\$\\sim\$\{\}/g;

  # conding spaces
  s/ /\~/g;

  # coding \"
  s/\@DQ\@/\\BS\{\}\"/g;

  # coding {}
  s/\@OB\@/\\\{/g;
  s/\@CB\@/\\\}/g;

  $_;
}

sub latex_symbols {
  local ($_) = @_;
  
  # coding backspace
  s/(\t| )*\\$/\@EOL\@/mg;
  s/\\/\@BS\@/g;
  
  # coding specials
  s/((\_|\&|\#|\$|\%))/\\$1/g;
  
  # coding ()[]{}
  s/(\{|\})/\\BRA\{\\$1\}/g;
  s/(\(|\)|\[|\])/\\BRA\{$1\}/g;

  # coding symbols
  s/\^/\$\\wedge\$\{\}/g;
  s/\~/\$\\sim\$\{\}/g;

  s/\.\*/\\DOTSTAR\{\}/g;
  s/-\>\*/\\PTRSTAR\{\}/g;
  s/-\>/\\PTR\{\}/g;
  
  s/\>\>/\\GG\{\}/g;
  s/\<\</\\LL{\}/g;
  s/\>=/\\GE\{\}/g;
  s/\<=/\\LE\{\}/g;
  s/\!=/\\NE\{\}/g;
  s/>/\\GT\{\}/g;
  s/</\\LT\{\}/g;

  s/\@DQ\@/\\BS\{\}\"/g;
  s/\@BS\@/\\BS\{\}/g;
  s/\@EOL\@/\\EOL\{\}/g;
  s/ /\~/g;

  $_;
}

sub do_spaces {
  local ($c) = @_;
  $c =~ s/ /s:/g;
  $c =~ s/\t/t:/g;
  my $cnt = 1;
  foreach $char (split(':',$c)) {
    if ( $char eq 't' ) {
      $cnt = $cnt + $tablen;
      $cnt = int ($cnt / $tablen);
      $cnt = $tablen * $cnt;
    } elsif ( $char eq 's' ) {
      $cnt++;
    }
  }
  my $spc = "";
  for ( my $i=1; $i < $cnt; $i++ ) {
    $spc .= "\~";
  }
  $spc;
}

sub add_alias {
  local ($a) = @_;
  $a =~ s/\@NL\@/ /g;
  $a =~ s/\\?\n/ /g;
  if ( $a =~ /\b(\w+)\s*\;/ ) {
    push(@ALIAS,$1);
  } elsif ( $a =~ /\b\w+\s+\(\s*\*?\s*(\w+)\s*\)/ ) {
    push(@ALIAS,$1);
  }
  "";
};

sub add_short_comment {
  local ($c) = @_;
  push (@SC,$c);
  "\@SC$#SC\@";
};

sub add_long_comment {
  local ($c) = @_;
  $c =~ s/\@SC(\d+)\@/$SC[$1]/g; # extract short comment
  push(@LC,$c);
  "\@LC$#LC\@";
}

sub add_line_comment {
  local ($c) = @_;
  $c =~ s/\@SC(\d+)\@/$SC[$1]/g; # extract short comment
  push(@CL,$c);
  "\@CL$#CL\@";
};

sub add_string {
  local ($c) = @_;
  push (@KEY,"\\CSTRING\{\"".&latex_chars($c)."\"\}");
  "\@KEY$#KEY\@";
};

sub add_string1 {
  local ($c) = @_;
  push (@KEY, "\\CSTRING\{".&latex_chars($c)."\}");
  "\@KEY$#KEY\@";
};

sub add_float {
  local ($a,$b) = @_;
  push (@KEY, "\\FLOAT\{".$a."\}\{".$b."\}");
  "\@KEY$#KEY\@";
}

sub add_number {
  local ($a) = @_;
  push (@KEY, "\\NUMBER\{".$a."\}");
  "\@KEY$#KEY\@";
}

sub add_begin_of_line {
  local ($a) = @_;
  push (@KEY, &do_spaces($a));
  "\@KEY$#KEY\@";
};

sub add_user {
  local ($a) = @_;
  push (@USER,$a);
  "\@USER$#USER\@";
}

sub add_macro {
  local ($a) = @_;
  push (@MACRO,$a);
  "\@MACRO$#MACRO\@";
}

sub add_preprocessing {
  local ($sa,$sb,$p) = @_;
  push (@KEY, &do_spaces($sa) . "\\PREPRO\{\\\#\}" .
              &do_spaces($sb) . "\\PREPRO\{$p\}");
  "\@KEY$#KEY\@";
}

sub add_keyword {
  local ($a) = @_;
  push (@KEY, "\\KEYWORD\{".&latex_chars($a)."\}" );
  "\@KEY$#KEY\@";
}

sub add_function {
  local ($a) = @_;
  push (@KEY, "\\FUNCTION\{".&latex_chars($a)."\}" );
  "\@KEY$#KEY\@";
}

sub do_encoding {

  local ($file) = @_;
  my $lines1;
  my $lines2;

  open (FILE, "<$file" );
  
  ############################################################################
  print "loading  file: ", $file, $endl unless $silent;
  $lines2 = "";
  while (<FILE>) {
    $lines2 .= $_;
  }
  close FILE;

  ############################################################################
  print "encoding short comments", $endl if $verbose;
  $lines2 =~ s/(\/\*.*?\*\/)/&add_short_comment($1)/mge;
  
  ############################################################################
  print "encoding multiple lines comment", $endl if $verbose;
  $lines2 =~ s/([^\/]|^)(\/\*.*?\*\/)/$1.&add_long_comment($2)/sge;
  
  ############################################################################
  print "encoding line comments", $endl if $verbose;
  $lines1 = "";
  foreach (split('\n',$lines2)) {
    while ( s/(\s+)\@LC(\d+)\@/\@LC$2\@/ ) {
      $LC[$2] = $1.$LC[$2];
    }
    s/(\/\/.*)$/&add_line_comment($1)/ge;
    $lines1 .= $_.$endl."\@NL\@".$endl;
  }
  $lines2 = $lines1;

  ###########################################################################
  print "encoding strings", $endl if $verbose;
  $lines2 =~ s/\\\"/\@DQ\@/sg;
  $lines2 =~ s/\"([^\"]*)\"/&add_string($1)/mge;
  $lines2 =~ s/(include\s+)(\<[a-zA-Z\_\/\.]+\>)/$1.&add_string1($2)/mge;

  ###########################################################################
  print "encoding numbers", $endl if $verbose;
  # number exponential form
  $lines2 =~ s/(\D)(\.\d+)\s*(e|E|d|D)\s*((\+|-)?\d+)/$1.&add_float($2,$4)/sge;
  $lines2 =~ s/\b(\d+\.?\d*)\s*(e|E|d|D)\s*((\+|-)?\d+)/&add_float($1,$3)/sge;
  $lines2 =~ s/(\D)(\.\d+)([^\@])/$1.&add_number($2).$3/sge;
  $lines2 =~ s/\b(\d+\.?\d*)([^\@])/&add_number($1).$2/sge;

  ###########################################################################
  print "encoding begin of line", $endl if $verbose;
  $lines2 =~ s/^(( |\t)+)/&add_begin_of_line($1)/mge;

  ############################################################################
  print "encoding namespaces, class, struct and union", $endl if $verbose;
  $lines2 =~ s/\b(namespace|class|struct|union)(\s*)\b(\w+)\b/$1.$2.&add_user($3)/sge;

  ############################################################################
  print "encoding alias", $endl if $verbose;
  $lines2 =~ s/\b(typedef)([^;]+;)/$1.$2.&add_alias($2)/sge;

  ###########################################################################
  print "encoding macros", $endl if $verbose;
  $lines2 =~ s/\b(define\s*)\b(\w+)\b/$1.&add_macro($2)/ge;

  ###########################################################################
  print "encoding preprocessing directives", $endl if $verbose;
  $lines2 =~ s/(\s*)\#(\s*)(($PREPROCESSING))\b/&add_preprocessing($1,$2,$3)/ge;

  ###########################################################################
  print "encoding keywords", $endl if $verbose;
  $lines2 =~ s/\b(($KEYWORDS))\b/&add_keyword($1)/sge;

  ###########################################################################
  print "encoding functions", $endl if $verbose;
  $lines2 =~ s/\b(($FUNCTIONS))\b/&add_function($1)/sge;

  ###########################################################################
  $FILES{$file} = $lines2;

}

sub do_more_encoding {

  local ($file) = @_;
  print "cross encoding file: ", $file, $endl unless $silent;
  
  my $tmp = $FILES{$file};
  my $i = 0;
  foreach my $a(@USER) {
    $tmp =~ s/\b$a\b/\@USER$i\@/sg;
    $i++;
  }
  $i = 0;
  foreach my $a(@ALIAS) {
    $tmp =~ s/\b$a\b/\@ALIAS$i\@/sg;
    $i++;
  }
  $i = 0;
  foreach my $a(@MACRO) {
    $tmp =~ s/\b$a\b/\@MACRO$i\@/sg;
    $i++;
  }

  $FILES{$file} = $tmp;
}


sub do_userkey_encoding {
        
  my $i;
  for ( $i = 0; $i <= $#LC; $i++ ) {
    $LC[$i] = "\\color{$comment_color}\n\\begin{verbatim}\n" .
              $LC[$i] .
              "\n\\end{verbatim}\n\\color{$text_color}";
  }
  for ( $i = 0; $i <= $#SC; $i++ ) {
    $SC[$i] = "\\COMMENT\{" .&latex_chars($SC[$i]) ."\}";
  }
  for ( $i = 0; $i <= $#CL; $i++ ) {
    $CL[$i] = "\\COMMENT\{" .&latex_chars($CL[$i]) ."\}";
  }
  for ( $i = 0; $i <= $#USER; $i++ ) {
    $USER[$i] = "\\CLASS\{".&latex_chars($USER[$i])."\}";
  }
  for ( $i = 0; $i <= $#ALIAS; $i++ ) {
    $ALIAS[$i] = "\\ALIAS\{".&latex_chars($ALIAS[$i])."\}";
  }
  for ( $i = 0; $i <= $#MACRO; $i++ ) {
    $MACRO[$i] = "\\MACRO\{".&latex_chars($MACRO[$i])."\}";
  }
}

sub do_decoding {

  local ($file) = @_;
  print "decoding file: ", $file, $endl unless $silent;
  
  my $lines = &latex_symbols($FILES{$file});

  $lines =~ s/\@KEY(\d+)\@/$KEY[$1]/sg;
  $lines =~ s/\@MACRO(\d+)\@/$MACRO[$1]/sg;
  $lines =~ s/\@USER(\d+)\@/$USER[$1]/sg;
  $lines =~ s/\@ALIAS(\d+)\@/$ALIAS[$1]/sg;
  $lines =~ s/\@SC(\d+)\@/$SC[$1]/sg;
  $lines =~ s/\@CL(\d+)\@/$CL[$1]/sg;
  $lines =~ s/\@LC(\d+)\@/$LC[$1]/sg;
  $lines =~ s/\@NL\@/\\par/sg;
  $lines =~ s/\n\n/\n\\\ \n/sg;

  open (OUT, ">$file.tex");
  print OUT &do_header($file), $lines, &do_footer, $endl;
  close OUT;

  my $dir;
  my $file_name;

  if ( $file =~ /(.*)\/(\w(\w|\.)+)/ ) {
    $dir       = $1;
    $file_name = $2;
  } else {
    $dir = ".";
    $file_name = $file;
  }

  my $dvips_cmd = "dvips";
  $dvips_cmd .= " -r"           if     $do_reverse;
  $dvips_cmd .= " -q"           unless $verbose;
  $dvips_cmd .= " -t landscape" if     $columns == 2;
  
  my $latex_cmd = "latex ${file_name}";
  $latex_cmd .= " > c++2ps.tmp; rm c++2ps.tmp" unless $verbose;

  my $rm_cmd = "rm -f ${file}.aux ${file}.dvi ${file}.log";
  $rm_cmd .= " ${file}.tex" unless $leave_tex;
  
  system("cd $dir; ${latex_cmd}; ${latex_cmd}");
  system("${dvips_cmd} ${file}.dvi -o ${file}.ps");
  system($rm_cmd);
}
